#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <string.h>

#if defined(__LINUX__)
#define DEV "/dev/sd%c"
#define START 'a'
#define END 'h'

#elif defined(__FREEBSD__)
#define DEV "/dev/da%i"
#define START 0
#define END 15

#elif defined(__OPENBSD__) || defined(__NETBSD__)
#define DEV "/dev/sd%i"
#define START 0
#define END 15
#endif

#define FIRMWARE_START 32256
#define BLOCK_SIZE 512
#define MAGIC_SIZE 8

#define NOIPOD 0
#define CAPPED 1
#define UNCAPPED 2

/* The iPod disk device */
int dev = 0;
/* Offset where the cap bytes are located on the current block */
int byte_cap = 0;
/* The buffer in which we store the data being read from the iPod */
char buffer [2*BLOCK_SIZE];
/* Magic string for 3G, 4G, mini, color iPods */
const char * magic1 = "\x08\x00\x9F\xE5\x38\x01\x90\xE5";
/* Magic string for video and video iPods */
const char * magic2 = "\x08\x00\x9F\xE5\x94\x00\x90\xE5";

/* 2 bytes telling if the iPod is capped or not */
const char * capped = "\x01\x00";
const char * uncapped = "\x00\x01";

/* Function that tries to detect the iPod (obvious isn't it ?) */
int detectIpod (void)
{
   char x;
   int j;

   for (x = START; x <= END; x++)
   {
      char devstring [10];
      sprintf (devstring, DEV, x);
      dev = open (devstring, O_RDWR);

      if (dev == -1)
         continue;

      lseek (dev, FIRMWARE_START, SEEK_SET);

      if (read (dev, buffer, BLOCK_SIZE) == -1)
         continue;

      for (j = 0; j < BLOCK_SIZE - 6; j++)
      {

         if (buffer [j] == 'S' && buffer [j+2] == 'T' && buffer [j+4] == 'O' && buffer [j+6] == 'P')
            return 1;
      }
      
      close(dev);
   }

   return 0;
}

/* Function that tries to locate a magic string that returns :
   - 0 if it didn't find it
   - 1 if it did and the iPod is capped
   - 2 if it did and the iPod is uncapped   
*/
int isCapped (void)
{
   int block = 0;
   
   if (read (dev, buffer + BLOCK_SIZE, BLOCK_SIZE) == -1)
         return -1;
   
   for (block = 0; block < 3000; block++)
   {
      if (memcpy (buffer, buffer + BLOCK_SIZE, BLOCK_SIZE) == NULL)
         return -1;
      
      if (read (dev, buffer + BLOCK_SIZE, BLOCK_SIZE) == -1)
         return -1;
      
      for (byte_cap = 0; byte_cap < BLOCK_SIZE; byte_cap++)
      {         
         if (memcmp (buffer + byte_cap, magic1, MAGIC_SIZE) == 0
             || memcmp (buffer + byte_cap, magic2, MAGIC_SIZE) == 0)
         {            
            if (memcmp (buffer + byte_cap + MAGIC_SIZE, capped, 2) == 0)
            {
               return CAPPED;
            }
            
            if (memcmp (buffer + byte_cap + MAGIC_SIZE, uncapped, 2) == 0)
            {
               return UNCAPPED;
            }
         }
      }
   }

   return NOIPOD;
}

/* Too obvious for you guys :-)
   Returns -1 if it failed, 0 if it succeeded.   
*/
int unCap (void)
{
   if (memcpy (buffer + byte_cap + MAGIC_SIZE, uncapped, 2) == NULL)
      return -1;
      
   if (lseek (dev, -(2*BLOCK_SIZE), SEEK_CUR) == -1)
      return -1;
   
   if (write (dev, buffer, BLOCK_SIZE) == -1)
      return -1;
      
   if (write (dev, buffer + BLOCK_SIZE, BLOCK_SIZE) == -1)
      return -1;
      
   return 0;
}

/* Same here */
int reCap (void)
{
   if (memcpy (buffer + byte_cap + MAGIC_SIZE, capped, 2) == NULL)
      return -1;
      
   if (lseek (dev, -(2*BLOCK_SIZE), SEEK_CUR) == -1)
      return -1;
   
   if (write (dev, buffer, BLOCK_SIZE) == -1)
      return -1;
      
   if (write (dev, buffer + BLOCK_SIZE, BLOCK_SIZE) == -1)
      return -1;
   
   return 0;
}


int main(int argc, char **argv)
{
   if (argv[1] == NULL || !strcmp(argv[1], "-h") || !strcmp(argv[1], "--help"))
   {
      printf ("goPod 1.1 by JiB, kang, Alf, g-rem & Darth Zgul\nUsage : gopod [-h] [-d] [-u] [-c]\n\n");
      printf ("-h\tdisplays this help\n-d\ttries to detect an iPod and tell if it is capped or not\n");
      printf ("-u\tuncaps your iPod\n-c\trecaps your iPod\n\nPlease use only one option at a time.\n");
      return 0;
   }

   if (strcmp (argv[1], "-d") == 0)
   {
      if (detectIpod())
      {
         int status = isCapped();

         if (status == -1)
         {
            fprintf (stderr, "Error: I/O Error while accessing to your iPod.\n");
         }
         
         if (status == NOIPOD)
         {
            fprintf (stderr, "Error: Unable to determine wether or not your iPod is capped.\n");
         }

         if (status == CAPPED)
         {
            printf ("Your iPod is capped. Use gopod -u to uncap it.\n");
         }
         
         if (status == UNCAPPED)
         {
            printf ("Your iPod isn't capped. Use gopod -c to recap it.\n");
         }         
      }

      else
      {
         printf ("No iPod detected.\n");
      }
   }

   if (strcmp (argv[1], "-u") == 0)
   {
      if (detectIpod() && isCapped() == CAPPED)
      {
         if (unCap() == -1)
            fprintf (stderr, "Error: Failed to uncap your iPod.\n");
      }

      else
      {
         fprintf (stderr, "Error: Failed to uncap your iPod.\n");
      }
   }

   if (strcmp (argv[1], "-c") == 0)
   {
      if (detectIpod() && isCapped() == UNCAPPED)
      {
         if (reCap() == -1)
            fprintf (stderr, "Error: Failed to recap your iPod.\n");
      }

      else
      {
         fprintf (stderr, "Error: Failed to recap your iPod.\n");
      }
   }

   close(dev);
   return 0;
}
